/*
 * anyRemote
 * a bluetooth remote for your PC.
 * Copyright (C) 2006
 * 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "parse.h"

#define ALARMS_SECTION_STR  "[Alarms]"
#define ALIAS_SECTION_STR   "[Aliases]"
#define KEYS_SECTION_STR    "[Keys]"
#define MODE_END_STR        "[ModeEnd]"
#define MODE_STR    	    "[Mode]"
#define SECTION_END_STR     "[End]"

#define AUTOCONN_STR   	"AutoConnect"
#define AUTOREPEAT_STR	"AutoRepeat"
#define BAUD_STR    	"Baudrate"
#define CHARSET_STR    	"CharSet"
#define CMER_OFF_STR   	"CmerOff"
#define CMER_ON_STR   	"CmerOn"
#define DEBUG_STR   	"Debug"
#define DEVICE_STR   	"Device"
#define LOG_STR    	"Log"
#define RETRY_STR    	"RetrySecs"
#define SCREEN_STR      "Screen"
#define TOMAIN_STR   	"ToMainMenu"

#define EMPTY_STR   	"EMPTY"

#define USEREGEX 1

#ifdef USEREGEX

#include <regex.h>

#define REGEX_KEYVAL   "^[[:space:]]*(\\[[[:alpha:]]+\\])([^=]*)$|^[[:space:]]*([^=%]*[^[:space:]]{1})[[:space:]]*=[[:space:]]*(.*[^[:space:]]{1})[[:space:]]*$"
#define REGEX_CMDBYCMD "[[:space:]]*(Exec|ExecAndSet|SetMode|Set|Send|ExecAndSend|Timer|CancelTimer|Macro)[[:space:]]*(\\()|[[:space:]]*(RemoteOn|RemoteOff|Exit|TempDisconnect)[[:space:]]*;{0,1}"

regex_t *keyVal   = NULL;
regex_t *cmdByCmd = NULL;

#endif

/* Data types */

mode     *modes       = NULL;
mode     *currentMode = NULL;
mode 	 *defMode     = NULL;
type_alias *aliases   = NULL;
type_key *alarms      = NULL;

typedef struct
{
	int log;
        int debug;
        int baudrate;
        int retrysecs;
        int screen;
        int autoConnect;               // AT mode only
        int autoRepeat;
        char charset   [MTEXTLEN];     // Charset - AT mode only
        char device    [MAXARGLEN];    // Device
        char toMain    [MTEXTLEN];     // AnyMenu  -> MainMenu - AT mode only
        char cmerOn    [MTEXTLEN];     // AT mode only
        char cmerOff   [MTEXTLEN];     // AT mode only
        int  model;
} CONF;

CONF conf = { 0, 1, B19200, 60, 0, 0, 0,
                "8859-1",               // charset
                DEFAULT_DEVICE,         // device
		"",   	    	    	// to main menu
    	    	"",     	    	// CMER ON
		"",     	    	// CMER OFF
		MODEL_DEFAULT
		};                         

/* ----------------- Interface functions ------------------------- */

int getLog()
{
    return conf.log;
}

int getDebug()
{
    return conf.debug;
}

char* getDevice()
{
    return conf.device;
}

int getBaudrate()
{
    return conf.baudrate;
}

int getRetrySecs()
{
    return conf.retrysecs;
}

int autoConnect()
{
    return conf.autoConnect;
}

int getAutoRepeat()
{
    return conf.autoRepeat;
}

int getUseScreen()
{
    return conf.screen;
}

char* getCharset()
{
    return (conf.charset[0] == '\0' ? NULL : conf.charset);
}

char* getToMainMenu()
{
    return  (conf.toMain[0] == '\0' ? NULL : conf.toMain);
}

char* getAT_CMER(int what)
{
    if (what == CMER_ON) {
    	if  (conf.cmerOn[0] == '\0') {
    	    if (conf.model == MODEL_MOTOROLA) {
	        return DEF_AT_CMER_ON_MOTOROLA;
    	    } else if (conf.model == MODEL_SE) {
	        return DEF_AT_CMER_ON_SE;
    	    } else if (conf.model == MODEL_SAGEM) {
	        return DEF_AT_CMER_ON_SAGEM;
	    } else if (conf.model == MODEL_SIEMENS) {
	    	return DEF_AT_CMER_ON_SIEMENS;
	    } else {
	        return DEF_AT_CMER_ON_DEFAULT;
	    }
        } else {
	    return conf.cmerOn;
	}
    }
	
    if (what == CMER_OFF) {
    	if  (conf.cmerOff[0] == '\0') {
    	    if (conf.model == MODEL_MOTOROLA) {
	        return DEF_AT_CMER_OFF_MOTOROLA;
    	    } else if (conf.model == MODEL_SE) {
	        return DEF_AT_CMER_OFF_SE;
    	    } else if (conf.model == MODEL_SAGEM) {
	        return DEF_AT_CMER_OFF_SAGEM;
	    } else if (conf.model == MODEL_SIEMENS) {
	    	return DEF_AT_CMER_OFF_SIEMENS;
	    } else {
	        return DEF_AT_CMER_OFF_DEFAULT;
	    }
        } else {
	    return conf.cmerOff;
	}
    }
    return NULL;
}

char* getAlarmInfo(int which, cmdItem ** cmds)
{
	//logger("DEBUG", "getAlarmInfo()");
	int i;
        type_key *tk = alarms;
        
	for (i=0;i<which;i++) {
        	if (tk != NULL) {
        		tk = tk->next;
                } else {
                	break;
                }
        }
        
        if (i<which || tk == NULL) {
		//logger("DEBUG", "getAlarmInfo() return NULL");
        	return NULL;
        }
    	*cmds = tk->cmd;
    
 	return tk->key;
}

void setMode (char *modeName) 
{
    if (modeName ==  NULL) {
        logger("DEBUG", "setMode() NULL input");
        return;
    }
    sprintf(tmp, "setMode() to %s", modeName);
    logger("DEBUG", tmp);

    mode *mp = modes;
    while (mp && strstr(modeName, mp->name) == NULL) {
        mp = (mode*) mp->next;
    }
    
    if (mp) {
        currentMode = mp;
        sprintf(tmp, "setMode() new mode was set to %s", currentMode->name);
        logger("DEBUG", tmp);
    } else {
        logger("DEBUG", "setMode() new mode did not found");
    }
    
    return;
}

char *getMode (void) 
{
    return currentMode->name;
}

void setModel (char *answer) 
{
    if (answer == NULL) {
    	conf.model = MODEL_DEFAULT;
    } else if (strstr(answer,STR_MOTOROLA)) {
    	conf.model = MODEL_MOTOROLA;
    } else if (strstr(answer,STR_SE)) {
    	conf.model = MODEL_SE;
    } else if (strstr(answer,STR_SAGEM)) {
    	conf.model = MODEL_SAGEM;
    } else if (strstr(answer,STR_SIEMENS)) {
    	conf.model = MODEL_SIEMENS;
    } else {
    	conf.model = MODEL_DEFAULT;
    }
}

int getModel (void) 
{
    return conf.model;
}

/* ----------------- Setup configuration functions ------------------------- */

static int setBaudrate(int rate) 
{
  switch (rate) {
    case 300:
        return 300;
    case 1200:
        return 1200;
    case 2400:
        return 2400;
    case 9600:
        return 9600;
    case 19200:
        return 19200;
    case 38400:
        return 38400;
#ifdef B57600
    case 57600:
        return 57600;
#endif
#ifdef B115200
    case 115200:
        return 115200;
#endif
#ifdef B230400
    case 230400:
        return 230400;
#endif
    default:
        printf("WARNING: bad baudrate %d, defaulting to 9600\n", rate);
  }

  return B9600;
}

/* Parses command line options and set flags. */
int parse_opts(int argc, char *argv[])
{
    int i;
    for(i=1; i<argc; i++)
    {
    	if(strcmp(argv[i],"-log")==0 || strcmp(argv[i],"-l")==0) 
	    conf.log=1;
    	else if((strcmp(argv[i],"--serial")==0 || strcmp(argv[i],"-s")==0) && i+1<argc) 
	    strcpy(conf.device, argv[++i]);
    	else if((strcmp(argv[i],"--baud")==0 || strcmp(argv[i],"-b")==0) && i+1<argc) 
	    conf.baudrate = setBaudrate(atoi(argv[++i]));
    	else if((strcmp(argv[i],"--noscreen")==0 || strcmp(argv[i],"-n")==0)) 
	    conf.screen = 0;
    	else if((strcmp(argv[i],"--autoconnect")==0 || strcmp(argv[i],"-a")==0)) 
	    conf.autoConnect = 1;
    	else if((strcmp(argv[i],"-f")==0)) {
	    ++i;
	    continue;	// already processed this parameter
    	} else if((strcmp(argv[i],"-h")==0) || (strcmp(argv[i],"--help")==0)) {
	    //Just print help and exit
	    printf("Usage: anyremote [-h] [-f conf.file] [-s serial] [-b baudrate] [-n] [-d]\n\n");
	    printf(" -h this help\n");
	    printf(" -f conf.file: specify non-default conf.file\n");
	    printf(" -s serial: specify serial to use\n");
	    printf(" -b baudrate: specify baudrate (default is 19200)\n");
	    printf(" -n: do not send any AT+CKPD command to the phone\n");
	    printf(" -a: reconnect automatically in case of connection failure\n\n");
	    exit(0);
    	} else {
    	    printf("ERROR: Unknown input parameter %s\n", argv[i]);
    	}
    }
    return 1;
}

// Returns true if the line can be ignored in the config file.
#ifndef USEREGEX
int ignoreline(char *line)
{
    if ((*line)==';' || (*line)=='%') return 1;
    while (isspace(*line) && *line) line++;
    return !(*line);
}
#endif

void stripHeadSpaces(char **str)
{
    if (*str != NULL) {
     	while (*str[0] != '\0' && (*str[0] == ' ' || *str[0] == '\t')) {
            (*str)++;
    	}
    }
}

void normalizeSequence(char **str)
{
    int isSpace = 0;
    
    char *n = (*str); 
    char *r = (*str);
    
    //printf("INFO : normalizeSequence in: >%s<\n", *str);
    
    while (*r != '\0') {
	if (isspace(*r)) {
	    // Skip spaces
	    if(!isSpace) {
		isSpace = 1;
	    }
	    r++;
	} else {
	    if(isSpace) {
	    	*n = ' ';
		n++;
	    }
	    // Copy all symbols
	    *n = *r;
	    r++;
	    n++;
	    isSpace = 0;
	}
    }
    *n = '\0';
    	
    //printf("INFO : normalizeSequence out: >%s<\n", *str);
}


#ifdef USEREGEX
int regexpPrepare(regex_t *r, char *p) 
{
    int err_no=0;

    if((err_no=regcomp(r, p, REG_EXTENDED))!=0) {
        size_t length; 
        char *buffer;
        length = regerror (err_no, r, NULL, 0);
        buffer = malloc(length);
        regerror (err_no, r, buffer, length);
        sprintf(tmp, "regcomp() error %s", buffer);
 
        free(buffer);
        regfree(r);
        return EXIT_ABORT;
    }

    return EXIT_OK;
}

int parseKeyValue(char *in, char ** tag, char ** value)
{
    //printf("INFO: parseKeyValue() %s (%d)\n", in, (int)keyVal->re_nsub+1);

    size_t no_sub = keyVal->re_nsub+1;    
    regmatch_t *result;
    if ((result = (regmatch_t *) malloc(sizeof(regmatch_t) * no_sub))==0) {
        printf("ERROR: No more memory");
        return EXIT_ABORT;
    }

    if (in[strlen(in) - 1] == '\n') {
        in[strlen(in) - 1] = '\0';
    }
    
    *value =  NULL;
    *tag   =  NULL;
    if (regexec(keyVal, in, no_sub, result, 0)==0) {

        int use1 = 1;
        int use2 = 2;
        if (result[1].rm_so == -1) {
            if (result[3].rm_so == -1) {
                printf("ERROR: Incorrectly formed command result[1/3].rm_so == -1\n");
                use1 = -1;
            }
            use1 = 3;
        } 
        if (result[2].rm_so == -1) {
            if (result[4].rm_so == -1) {
                printf("ERROR: Incorrectly formed command result[2/4].rm_eo == -1\n");
                use2 = -1;
            }
            use2 = 4;
        } 

        if (use1 > 0) {
            *tag   = in + result[use1].rm_so;
            *(in + result[use1].rm_eo) = '\0';
        }
        if (use2 > 0) {
            if (result[use2].rm_so == -1 || result[use2].rm_eo == -1) {
                *value =  NULL;
            } else {  
                *value = in + result[use2].rm_so;
                *(in + result[use2].rm_eo) = '\0';
            }
        }
        /*
	if(*tag != NULL) {
            printf("INFO : Got tag   : >%s<\n", *tag);
        }else {
            printf("INFO : Got tag : >NULL<\n");
        }
        if(*value != NULL) {
    	    printf("INFO : Got value : >%s<\n", *value);
        }else {
            printf("INFO : Got value : >NULL<\n");
        }
	*/
    //} else {
    //    printf("INFO : Not matched\n");
    }

    free(result);
    return EXIT_OK;
}
#else
void parseKeyValue(char *in, char ** tag, char ** value)
{
    char *ptr;
    //printf("INFO : parseKeyValue()\n");
    
    ptr = (char *) in + strlen(in) - 1;
    
    // strip spaces from tail
    while( *ptr == '\0' || *ptr == '\n' || *ptr == '\t' || *ptr == ' ') {
    	*ptr = '\0';
        ptr--;
    }
    
    ptr = in;
    if (*ptr == '\\')  {     // Screened symbol
    	ptr++;
    }
    *tag = ptr;
    
    while (*ptr != '\0' && *ptr != '=') {
        ptr++;
    }
    if (*ptr == '\0') {
        //printf("WARNING : parseKeyValue: value is null for tag %s\n", *tag);
        return;
    }
    *ptr = '\0';
    ptr++;
    *value = ptr;
    stripHeadSpaces(value);
   
    ptr = in;
    
    stripHeadSpaces(tag);
    
    ptr = ptr + strlen(ptr) - 1;
    // strip spaces from tail
    while( *ptr == '\0' || *ptr == '\n' || *ptr == '\t' || *ptr == ' ') {
    	*ptr = '\0';
        ptr--;
    }
/*
    printf("INFO : Got tag   : >%s<\n", *tag);
    if(*value) {
    	printf("INFO : Got value : >%s<\n", *value);
    }else {
        printf("INFO : Got value : >NULL<\n");
    }
*/
    return;
}
#endif

void printParams()
{
    int atMode = 1;
    if (strstr(conf.device, INET_SOCKET) != NULL || 
       strstr(conf.device, BT_SOCKET) != NULL) {
        atMode = 0;
    }

    logger("CFG", "[Parameters]");
    if (atMode) {
        sprintf(tmp, "\tAutoConnect   %s", (conf.autoConnect?"Yes":"No"));
        logger("CFG", tmp);
    }
    sprintf(tmp, "\tAutoRepeat    %s", (conf.autoRepeat?"Yes":"No"));
    logger("CFG", tmp);
    if (atMode) {
	if (conf.cmerOn[0] != '\0') {
	    sprintf(tmp, "\tCMER On       %s", conf.cmerOn);
	    logger("CFG", tmp);
	} else {
	    logger("CFG", "\tCMER On       Will be determined automatically");
	}
	if (conf.cmerOff[0] != '\0') {
	    sprintf(tmp, "\tCMER Off      %s", conf.cmerOff);
	    logger("CFG", tmp);
	} else {
	    logger("CFG", "\tCMER Off      Will be determined automatically");	
	}
    }
    sprintf(tmp, "\tDebug         %s", (conf.debug?"Yes":"No"));
    logger("CFG", tmp);
    sprintf(tmp, "\tDevice        %s", conf.device);
    logger("CFG", tmp);
    sprintf(tmp, "\tLog           %s", (conf.log?"Yes":"No"));
    logger("CFG", tmp);
    sprintf(tmp, "\tRetry Sec     %d", conf.retrysecs);
    logger("CFG", tmp);
    sprintf(tmp, "\tScreen Menu   %s", (conf.screen?"Yes":"No"));
    logger("CFG", tmp);
    if (atMode) {
	sprintf(tmp, "\tTo Main Menu  %s", conf.toMain);
	logger("CFG", tmp);
    }
    logger("CFG", SECTION_END_STR);
} 

static char* id2Cmd (int cmdId) {
    switch(cmdId) {
        case ID_EXIT:               return CMD_EXIT;
        case ID_TMPDISCONN:         return CMD_TMPDISCONN;
        case ID_REMOTEON:           return CMD_REMOTEON;
        case ID_REMOTEOFF:          return CMD_REMOTEOFF;
        case ID_EXEC:               return CMD_EXEC;
        case ID_SENDCKPD:           return CMD_SENDCKPD;
        case ID_SETMODE:            return CMD_SETMODE;
        case ID_SET:                return CMD_SET;
        case ID_EXECSET:            return CMD_EXECSET;
        case ID_TIMERCMD:           return CMD_TIMER;
        case ID_SEND:               return CMD_SEND;
        case ID_EXECSEND:           return CMD_EXECSEND;
        case ID_CANCELTIMER:        return CMD_CANCELTIMER;
        case ID_MACRO:              return CMD_MACRO;
        
        default:                    return "Unknown";
    }
}

void printKeysInMode(type_key* It)
{    
        while (It && It->key != NULL) {
            if (It->cmd) {
    	        sprintf(tmp, "\t%s = ", It->key);
                
                cmdItem* item = It->cmd;
                while (item) { 
                    
                    strcat(tmp, id2Cmd(item->type));
                    if (item->descr) {
                        strcat(tmp, "(");
                        strcat(tmp, item->descr);
                        strcat(tmp, ")");
                    }
                    strcat(tmp, ";");
                    
                    item = item->next;
                }
                
    	    } else {
                sprintf(tmp, "\t%s = no command", It->key);
            }    
    	    logger("CFG", tmp);
	
	    It = (type_key*)It->next;
        }
        return;
}

void printKeys()
{    
    mode *mp = modes;
    while (mp) {
    
        sprintf(tmp, "%s %s", MODE_STR, mp->name);
        logger("CFG", tmp);
         
        type_key* It = mp->keys;
        
        printKeysInMode(It);
    
        sprintf(tmp, "%s %s", MODE_END_STR, mp->name);
        logger("CFG", tmp);
        
        mp = (mode*)mp->next;
    }
    
    return;
}

void printAlarms()
{
        logger("CFG", ALARMS_SECTION_STR);

	printKeysInMode(alarms);

        logger("CFG", SECTION_END_STR);
	return;
}

void printAliases()
{
    type_alias *al = aliases;

    logger("CFG", ALIAS_SECTION_STR);

    while(al) {
    	sprintf(tmp, "\tKey %s \taliased to %s", al->key, al->alias);
    	logger("CFG", tmp);

        al = al->next;
    }
    logger("CFG", SECTION_END_STR);

    return;
}

#ifdef USEREGEX
int storeCmds(cmdItem** cmdList, char *value) 
{
    cmdItem *curCmd = (cmdItem*) NULL;
    int cmdId;
    
    size_t no_sub = cmdByCmd->re_nsub+1;    
    //printf("storeCmds() %s (%d)\n", value, (int)no_sub); 
    regmatch_t *result;
    if ((result = (regmatch_t *) malloc(sizeof(regmatch_t) * no_sub))==0) {
        printf("ERROR: No more memory");
        return EXIT_ABORT;
    }
 
    int start = 0;
    while(regexec(cmdByCmd, value+start, no_sub, result, 0)==0) {
        
        //printf("storeCmds() next %s\n", value+start); 
    
        int  step  = 0;
        char *name = NULL;
        char *cmds = NULL;
        
        int u1 = 1;
        if (result[1].rm_so >= 0 && result[2].rm_so) {  // <name>(<params>) matched  
            cmds = value + start + result[2].rm_so + 1;
             
            // Search ) which is appropriate to matched (
            int braces = 1;         // cmd_name( 
            char *ptr = cmds;         
            while (*ptr != '\0' && braces > 0) {
                if (*ptr == '(') {
                    braces++;
                }
                if (*ptr == ')') {
                    braces--;
                }
                ptr++;
            } 
            step = ptr - value - start + 1;
            if (*(ptr-1) == ')') {
                *(ptr-1) = '\0';
            }
        } else if (result[3].rm_so >= 0) {  // RemoteOff, RemoteOn, Exit, TempDisconnect matched
            u1 = 3;
            step = result->rm_eo;        
        } else {
            printf("Strange match\n");
            start += result->rm_eo;        
            continue;
        }
        
        name   = value + start + result[u1].rm_so;
        *(value + start + result[u1].rm_eo) = '\0';
        
        if (name == NULL) {
            //printf("Got name >NULL<\n");
            start +=result->rm_eo;        
            continue;
        //} else {
        //    printf("Got name >%s<\n",name);
        }
        
        if (strcmp(name,CMD_EXIT) == 0) {
            cmdId = ID_EXIT;
        } else if (strcmp(name,CMD_TMPDISCONN) == 0) {
            cmdId = ID_TMPDISCONN;
        } else if (strcmp(name,CMD_REMOTEON) == 0) {
            cmdId = ID_REMOTEON;
        } else if (strcmp(name,CMD_REMOTEOFF) == 0) {
            cmdId = ID_REMOTEOFF;
        } else if (strcmp(name,CMD_EXEC) == 0) {
            cmdId = ID_EXEC;
        } else if (strcmp(name,CMD_SENDCKPD) == 0) {
            cmdId = ID_SENDCKPD;
        } else if (strcmp(name,CMD_SETMODE) == 0) {
            cmdId = ID_SETMODE;
        } else if (strcmp(name,CMD_SET) == 0) {
            cmdId = ID_SET;
        } else if (strcmp(name,CMD_EXECSET) == 0) {
            cmdId = ID_EXECSET;
        } else if (strcmp(name,CMD_TIMER) == 0) {
            cmdId = ID_TIMERCMD;
        } else if (strcmp(name,CMD_SEND) == 0) {
            cmdId = ID_SEND;
        } else if (strcmp(name,CMD_EXECSEND) == 0) {
            cmdId = ID_EXECSEND;
        } else if (strcmp(name,CMD_CANCELTIMER) == 0) {
            cmdId = ID_CANCELTIMER;
        } else if (strcmp(name,CMD_MACRO) == 0) {
            cmdId = ID_MACRO;
        } else {
            printf("WARNING: Unknown command name\n");
            //here could be params like log=... also
            cmdId = ID_UNKNOWN;
            start +=result->rm_eo;        
            continue;
        }
        
        /*
	if (cmds != NULL) {
            printf("Got command body >%s<\n",cmds);
        } else {
            printf("Got command body >NULL<\n");
        }
	*/
        //printf("the rest of cmd >%s<\n",value+start+step);
        
        // Insert into command list
        cmdItem* newCmd = calloc(1, sizeof(cmdItem));
        
        newCmd->type = cmdId;
        newCmd->next = NULL;
        if (cmds == NULL) {
            newCmd->descr = NULL;
        } else {
            char *newDescr = calloc(1, strlen(cmds) + 1);
            strcpy(newDescr, cmds);
            newCmd->descr = newDescr;
        }
        
        if (curCmd == NULL) { // Insert first
            *cmdList = newCmd;
        } else {
            curCmd->next = newCmd;
        }
        curCmd = newCmd;

        start +=step;
        //printf("REST OF CMDS %s\n", value+start);        
    }   // while
    free(result);
    
    return EXIT_OK;
}
#else
int storeCmds(cmdItem** cmdList, char *value) 
{
    
    char name[MTEXTLEN];
    char body[MAXCMDLEN];
    char *ptr;
    cmdItem *curCmd;
    int idx, cmdId, skipToSemicolon;
    
    if (value == NULL) {
        printf("storeCmds(): null input\n"); 
        return EXIT_OK;
    }
    //printf("storeCmds() %s\n", value); 
     
    curCmd = (cmdItem*) NULL;
    ptr = value;
    
    while (*ptr != '\0') {
        //printf("Full line >%s<\n",ptr);
        
        idx = 0;
        skipToSemicolon = 0;
        
        while (*ptr != '\0' && (*ptr == ' ' || *ptr == '\t')) {
            ptr++;
        }
	
        // Get name
        while (*ptr != '\0' && *ptr != '(' && *ptr != ';' && idx < MTEXTLEN) {
            
            name[idx] = *ptr;
            idx++;
            ptr++;
        }
	if (idx >= MTEXTLEN) {
	    printf("Command >%s< is incorrect\n", value); 
	    return EXIT_NOK;
	}
	while (name[idx-1] == ' ' || name[idx-1] == '\t') {    // go backward to strip spaces 
                idx--;
        }
        name[idx] = '\0';
        
        //printf("Got name >%s<\n",name);
        
        if (strcmp(name,CMD_EXIT) == 0) {
            cmdId = ID_EXIT;
            skipToSemicolon = 1;
        } else if (strcmp(name,CMD_TMPDISCONN) == 0) {
            cmdId = ID_TMPDISCONN;
            skipToSemicolon = 1;
        } else if (strcmp(name,CMD_REMOTEON) == 0) {
            cmdId = ID_REMOTEON;
            skipToSemicolon = 1;
        } else if (strcmp(name,CMD_REMOTEOFF) == 0) {
            cmdId = ID_REMOTEOFF;
            skipToSemicolon = 1;
        } else if (strcmp(name,CMD_EXEC) == 0) {
            cmdId = ID_EXEC;
        } else if (strcmp(name,CMD_SENDCKPD) == 0) {
            cmdId = ID_SENDCKPD;
        } else if (strcmp(name,CMD_SETMODE) == 0) {
            cmdId = ID_SETMODE;
        } else if (strcmp(name,CMD_SET) == 0) {
            cmdId = ID_SET;
        } else if (strcmp(name,CMD_EXECSET) == 0) {
            cmdId = ID_EXECSET;
        } else if (strcmp(name,CMD_TIMER) == 0) {
            cmdId = ID_TIMERCMD;
        } else if (strcmp(name,CMD_SEND) == 0) {
            cmdId = ID_SEND;
        } else if (strcmp(name,CMD_EXECSEND) == 0) {
            cmdId = ID_EXECSEND;
        } else if (strcmp(name,CMD_CANCELTIMER) == 0) {
            cmdId = ID_CANCELTIMER;
        } else if (strcmp(name,CMD_MACRO) == 0) {
            cmdId = ID_MACRO;
        } else {
            //printf("WARNING: Unknown command name\n");
            //here could be params like log=... also
            cmdId = ID_UNKNOWN;
            skipToSemicolon = 1;
        }

	if (skipToSemicolon) {      // Skip until next ';'
            while (*ptr != '\0' && *ptr != ';') {
                ptr++;
            }
            if (*ptr != '\0') {
                 ptr++;
            }
            if (cmdId == ID_UNKNOWN) {
                continue;               // Next loop - read command by command
            }
        }
     
        // skip possible spaces
        while (*ptr != '\0' && (*ptr == ' ' || *ptr == '\t')) {
            ptr++;
        }
    
        // We've got name of command, now we have to get body
        if (skipToSemicolon) { // *ptr == ';' || *ptr == '\0') 	// Command without body 
            body[0] = '\0';
        } else {
            
            idx = 0;
            while (*ptr != '\0' && *ptr != '(') {
                ptr++;
            }
            if (*ptr == '(') {
                ptr++;
            }
            int braces = 1;         // cmd_name(          
            while (*ptr != '\0' && !(*ptr != ';' && braces == 0)) {
                body[idx] = *ptr;
                if (*ptr == '(') {
                    braces++;
                }
                if (*ptr == ')') {
                    braces--;
                }
                
                idx++;
                ptr++;
            } 
            
            while (idx > 0 && body[idx] != ')') {    // go backward to strip last ')' 
                idx--;
            }
            body[idx] = '\0'; // instead of ')'
        }
        //printf("Got command body >%s<\n",body);
        //printf("the rest of cmd >%s<\n",ptr);
        
        // Insert into command list
        cmdItem* newCmd = calloc(1, sizeof(cmdItem));
        
        newCmd->type = cmdId;
        newCmd->next = NULL;
        if (body[0] == '\0') {
            newCmd->descr = NULL;
        } else {
            char *newDescr = calloc(1, strlen(body) + 1);
            strcpy(newDescr, body);
            body[0] = '\0';
            newCmd->descr = newDescr;
        }
        
        if (curCmd == NULL) { // Insert first
            *cmdList = newCmd;
        } else {
            curCmd->next = newCmd;
        }
        curCmd = newCmd;
        
    }   // while
    
    return EXIT_OK;
}
#endif

static int storeKey(type_key **head, char *tag, char *value) 
{
	type_key* It = NULL;
	int ret;

	normalizeSequence(&tag);

	It = (type_key*) calloc(1, sizeof(type_key));

	// Insert in head
	if ((*head) == NULL) {   // first elem
	 	It->next = NULL;
	} else {
		It->next = (type_key*) (*head);
	}
	*head = It;
	
        It->key = (char*) calloc(1, strlen(tag)+1);
	strcpy(It->key,tag);
	if (value!= NULL) {
		ret = storeCmds(&It->cmd, value);
	} else {
		It->cmd = NULL;
	}

	return EXIT_OK;
}


int loadKeys(FILE *fp)
{
	char tmptext[MAXLEN]; 
	char *ret;
	char *tag, *value;    

	// We have at least default mode
	defMode = (mode*) calloc(1, sizeof(mode));
	strcpy(defMode->name,"default");
	defMode->next = NULL;
	defMode->keys = NULL;

	modes	      = defMode;
	currentMode   = defMode;
	mode *curMode = modes;

	do {
		ret=fgets(tmptext, MAXLEN, fp);
	
		if (ret == NULL) return EXIT_NOK;
		#ifndef USEREGEX
		if (ignoreline(tmptext)) continue;
		#endif
	
		tag   = NULL;
		value = NULL;
		parseKeyValue(tmptext, &tag, &value);
	
		if (tag == NULL) { 
			//printf("ERROR : incorrect command %s\n", tmptext);
			continue;
		} else if(strcmp(tag,SECTION_END_STR) == 0) {
			break;
		} else if(strcmp(tag,MODE_STR) == 0) { // mode definition
			mode *nm = (mode*) calloc(1, sizeof(mode));
			strcpy(nm->name,value);
			nm->keys = NULL;
			    
			// Insert in head
			if (modes == NULL) {   // first elem
				nm->next = NULL;
			} else {
				nm->next = (mode*) modes;
			}
			modes	= nm;
			curMode = nm;
	
		} else if(strcmp(tag,MODE_END_STR) == 0) { // mode definition end
			curMode = defMode;
		} else if (tag != NULL) {	    // Line with key->command definition
			storeKey(&(curMode->keys), tag, value);
		} else {
			continue;
		}
	} while (1);

	return EXIT_OK;    
}

int loadAliases(FILE *fp)
{
	char tmptext[MAXLEN]; 
	char *ret;
	char *key, *alias;
	    

	do {
		ret=fgets(tmptext, MAXLEN, fp);
	
		if (ret == NULL) return EXIT_NOK;
		#ifndef USEREGEX
		if (ignoreline(tmptext)) continue;
		#endif
	
		key   = NULL;
		alias = NULL;
		parseKeyValue(tmptext, &key, &alias);
	
		if (key == NULL) { 
			//printf("ERROR : incorrect alias %s\n", tmptext);
			continue;
		} else if(strcmp(key,SECTION_END_STR) == 0) {
			break;
		} else if (alias != NULL) {	// Line with key=alias definition
			
			type_alias* Al = (type_alias*) calloc(1, sizeof(type_alias));

			// Insert in head
			if (aliases == NULL) {   // first elem
				Al->next = NULL;
			} else {
				Al->next = (type_alias*) aliases;
			}
			aliases = Al;

			strcpy(Al->key,  key);
			strcpy(Al->alias,alias);
		} else {
			continue;
		}
	} while (1);

	return EXIT_OK;    
}

int loadAlarms(FILE *fp)
{
	char tmptext[MAXLEN]; 
	char *ret;
	char *tag, *value;    

	do {
		ret=fgets(tmptext, MAXLEN, fp);
	
		if (ret == NULL) return EXIT_NOK;
		#ifndef USEREGEX
		if (ignoreline(tmptext)) continue;
		#endif
	
		parseKeyValue(tmptext, &tag, &value);
	
		if (tag == NULL) { 
			continue;
		} else if(strcmp(tag,SECTION_END_STR) == 0) {
			break;
		} else if (value != NULL) {// Line with key->command definition
			storeKey(&alarms, tag, value);
		} else {
			continue;
		}
	} while (1);

	return EXIT_OK;    
}

int load_cfg(char *mfile)
{
	FILE *fp;
	char tmptext[MAXLEN], *ret;
	char *tag, *value;

	fp=fopen(mfile,"r");
	if (fp == NULL) return -1;

	#ifdef USEREGEX
	keyVal  = (regex_t *) malloc(sizeof(regex_t));
	memset(keyVal,   0, sizeof(regex_t));
	cmdByCmd = (regex_t *) malloc(sizeof(regex_t));
	memset(cmdByCmd, 0, sizeof(regex_t));

	if (regexpPrepare(keyVal,   REGEX_KEYVAL)   == EXIT_ABORT ||
		regexpPrepare(cmdByCmd, REGEX_CMDBYCMD) == EXIT_ABORT) {
		exit(1);
	}
	#endif

	// Go through the conf file
	while(1) {
		
		ret=fgets(tmptext, MAXLEN, fp);

		if (ret==NULL) break;
		#ifndef USEREGEX
		if (ignoreline(tmptext)) continue;
		#endif
		    	
		parseKeyValue(tmptext, &tag, &value);
			
		if (tag == NULL) { 
			//printf("INFO : skip line %s\n", tmptext);
			continue;
		} else if (strcmp(tag,RETRY_STR) == 0) {
			if (value != NULL) {
				conf.retrysecs=atoi(value);
			}	    
		} else if (strcmp(tag,DEVICE_STR) == 0) {
			if (value != NULL) {
				strcpy(conf.device,value);
			}	    
		} else if (strcmp(tag,AUTOCONN_STR) == 0) {
			if (value != NULL && (strncmp("Yes",value,3)==0 || strncmp("true",value,4)==0)) {
			 	conf.autoConnect=1;
			}	    
		} else if (strcmp(tag,CHARSET_STR) == 0) {
		    if (value != NULL) {
		    	strcpy(conf.charset,value);
		    }		
		} else if (strcmp(tag,LOG_STR) == 0) {
			if (value != NULL && (strncmp("Yes",value,3)==0 || strncmp("true",value,4)==0)) {
				conf.log=1;
			}	    
		} else if (strcmp(tag,SCREEN_STR) == 0) {
			if (value != NULL && (strncmp("Yes",value,3)==0 || strncmp("true",value,4)==0)) {
				conf.screen=1;
			}	    
		} else if (strcmp(tag,BAUD_STR) == 0) {
			if (value != NULL) {
				conf.baudrate = setBaudrate(atoi(value));
			}	    
		} else if (strcmp(tag,KEYS_SECTION_STR) == 0) {
			if (loadKeys(fp) != EXIT_OK) {
				return EXIT_NOK;
			} 
		} else if (strcmp(tag,ALIAS_SECTION_STR) == 0) {
			if (loadAliases(fp) != EXIT_OK) {
				return EXIT_NOK;
			}
		} else if (strcmp(tag,ALARMS_SECTION_STR) == 0) {
			if (loadAlarms(fp) != EXIT_OK) {
				return EXIT_NOK;
			} 
		} else if (strcmp(tag,TOMAIN_STR) == 0) {
			if(value!= NULL) {
				strcpy(conf.toMain,value);
			}	    
		} else if (strcmp(tag,DEBUG_STR) == 0) {
			if (value != NULL && (strncmp("No",value,3)==0 || strncmp("false",value,4)==0)) {
				conf.debug=0;
			}	    
		} else if (strcmp(tag,AUTOREPEAT_STR) == 0) {
			if (value != NULL && (strncmp("Yes",value,3)==0 || strncmp("true",value,4)==0)) {
				conf.autoRepeat=1;
			}	    
		} else if (strcmp(tag,CMER_ON_STR) == 0) {
			if(value!= NULL) {
				strcpy(conf.cmerOn,value);
			}	    
		} else if (strcmp(tag,CMER_OFF_STR) == 0) {
			if(value!= NULL) {
				strcpy(conf.cmerOff,value);
			}	    
		} else {
			printf("WARNING: Unknown tag in cfg file >%s<\n",tag);
		}
	}

	fclose(fp);

	return 1;
}

void printConf() 
{
	printKeys   ();
	printAliases();
	printAlarms ();
	printParams ();

	return ;
}

int init_cfg(char *path)
{
	if (path != NULL) {
		
		printf("CFG: Use cfg file >%s<\n",path);
		return load_cfg(path);
	} else {
		char cfgfile[MAXLEN];

		strcpy(cfgfile, getenv("HOME"));
		strcat(cfgfile, CFGFILE);

		printf("CFG: Use cfg file %s\n", cfgfile);
		if(load_cfg(cfgfile) != 1) {
			// Try again in current dir 
			strcpy(cfgfile, ".");
			strcat(cfgfile, CFGFILE);
			
			printf("CFG: Use cfg file %s\n", cfgfile);
			return load_cfg(cfgfile);
		}
	}
	return 1;
}

void freeCfg(void)
{
	logger("DEBUG", "freeCfg()");

	// Keys
	while(modes) {
		mode	 *md = modes;
		type_key *mk = md->keys;
	
		while (mk) {
			type_key* It = mk;
			mk = (type_key*) mk->next;
	
			cmdItem* item = It->cmd;
			cmdItem* mi;
			while (item) {
				mi = item->next;
				if (item->descr != NULL) {
					free(item->descr);
					item->descr = NULL;
				}
				free(item);
				item = mi;
			}
			It->cmd = NULL;
                        free(It->key);
			free(It);
		}
		modes = (mode*) modes->next;
		free(md);
	}

	// Alarms
        type_key *mk = alarms;
	while (mk) {
		type_key* It = mk;
		mk = (type_key*) mk->next;
	
		cmdItem* item = It->cmd;
		cmdItem* mi;
		while (item) {
			mi = item->next;
			if (item->descr != NULL) {
				free(item->descr);
				item->descr = NULL;
			}
			free(item);
			item = mi;
		}
		It->cmd = NULL;
                free(It->key);
		free(It);
	}

	// Aliases
	while(aliases) {
		type_alias *al = aliases->next;
		free(aliases); aliases = NULL;
		aliases = al;
	}

	logger("DEBUG", "freeCfg() OK");
}

/* ------------------------- Utility finctions -------------------------- */

static type_key* findExact(mode *mode, char *key)
{
	type_key* It = mode->keys;

	// Search exact command
	while (It && It->key != NULL  && strcmp(It->key,key) != 0) {
		//sprintf(tmp,"findExact search >%s< compare to >%s<", It->key, key);
		//logger("DEBUG",tmp);   

		It = (type_key*) It->next;
	}
	return It;
}

static type_key* findStartingWith(mode *mode, char *key)
{
	type_key* It = mode->keys;

	while (It && It->key != NULL) {
		//sprintf(tmp,"findStartingWith search >%s< compare to >%s<", It->key, key);
		//logger("DEBUG",tmp);   
		char *start = strstr(It->key,key);

		if (start && start == It->key) {    // We got a part of multi-key or parametrized command
			
			//logger("DEBUG","Found part of multi-key or parametrized command");
			return It;
		} 

		It = (type_key*) It->next;
	}    

	return It;
}


static type_key* findItemInMode(mode *mode, char *key, int *flag, cmdParams *params) 
{
	if (mode == NULL) {
		logger("DEBUG","findItemInMode: mode is null");
		return NULL;
	}

	sprintf(tmp,"findItemInMode >%s,%s<", mode->name, key);
	logger("DEBUG",tmp);

	// Clean-up
	if (params != NULL) {
		params->index[0] = '\0';
		params->value[0] = '\0';
	}

	type_key* It = findExact(mode,key);	// Search exact command

	if (It) {
		//logger("DEBUG","Found exact command");
		*flag = FLAG_EXACT;
		return It;
	} 

	It = findStartingWith(mode,key);	// Search as part of multikey sequence
	if (It) {
		//logger("DEBUG","Found part of multi-key command");
		*flag = FLAG_MULTIKEY;
		return It;
	} 

	if (params == NULL) {
		logger("DEBUG","findItemInMode: No parameters suspected. Item does not found.");
		return NULL;
	}

	// Search as parametrized command
	char *start  = index(key,'(');
	char *finish = rindex(key,')');

	if (finish == NULL) {  // Command was read partially ?
		finish = key + strlen(key);
	}

	if (start != NULL && finish != NULL && start != key) {        // Do not match "(...)=..."
		It = mode->keys;
	
		char tag  [MAXARGLEN];
		char index[4];
		char value[MAXARGLEN];
	
		strncpy(tag, key, start-key);
		*(tag+(start-key)) = '\0';
	
		char *comma = strstr(start+1,",");
		if (comma == NULL || comma > finish) { // Parametrized command from Java Client should be in a form like List(1,String1)
			logger("DEBUG","findItemInMode: Incorrectly formed parametrized command");
			return NULL;
		}
		strncpy(index,start+1,comma-start-1);
		*(index+(comma-start-1)) = '\0';

		strncpy(value,comma+1,finish-comma-1);
		*(value+(finish-comma-1)) = '\0';

		sprintf(tmp,"Parametrized command parsed as >%s,%s,%s< ", tag,index,value);
		logger("DEBUG",tmp);
	
		// Try to search explicitly specified parametrized command (for ex List(1) or List(commandX))
		// By index
		strcat(tag,"(");
		strcat(tag,index);
		strcat(tag,")");
		It = findExact(mode,tag);

		if (It) {
			logger("DEBUG","Found exact (explicitly specified by index) parametrized command");
			*flag = FLAG_EXACT;
			return It;
		} 

		// Then by value
		*(tag+(start-key+1)) = '\0';
		strcat(tag,value);
		strcat(tag,")");
		It = findExact(mode,tag);

		if (It) {
			logger("DEBUG","Found exact (explicitly specified by value) parametrized command");
			*flag = FLAG_EXACT;
			return It;
		} 
	
		// Finally... Search command like List($(Something))
		*(tag+(start-key)) = '\0';
		It = findStartingWith(mode,tag);
		if (It) {
			logger("DEBUG","Found parametrized command");

			strcpy(params->index, index);
			strcpy(params->value, value);
	
			*flag = FLAG_PARAMETR;
			return It;
		} 
	}

	logger("DEBUG","findItemInMode: Item does not found");
	return NULL;
}

char* findAlias(char *keyIn)
{
	//logger("DEBUG","findAlias");
	type_alias* al = aliases;
	char key[MAXARGLEN];

	if (*keyIn == '\0') {
		//logger("INFO","findAlias: Got empty key");   
		strcpy(key, EMPTY_STR);
	} else {
	 	strncpy(key, keyIn, MAXARGLEN - 1);
	}

	while (al) {
		//sprintf(tmp,"compare alias >%s< = >%s<", key, al->key);
		//logger("DEBUG",tmp);   

		if (strcmp(key, al->key) == 0) {
			sprintf(tmp,"found alias %s=%s", key, al->alias);
			logger("DEBUG",tmp);   
	
			return al->alias;
		}
		al = al->next;
	}
	return NULL;
}

type_key* findItem(char *keyIn, int *flag, cmdParams *params) 
{
	logger("DEBUG","findItem()"); 
	type_key *tk = NULL;

	char key[MAXARGLEN];

	if (*keyIn == '\0') {
		logger("INFO","Got empty CKEV");   
		strcpy(key, EMPTY_STR);
	} else {
		strncpy(key, keyIn, MAXARGLEN - 1);
	}
	 
	if (currentMode != NULL) {
		tk = findItemInMode(currentMode, key, flag, params);
		if (tk == NULL && currentMode != defMode) {
			logger("DEBUG","findItem() search in default");
			tk = findItemInMode(defMode, key, flag, params);
		}
	}
	return tk;
}

cmdItem *getCommand(type_key* item)
{
	if (!item) {
		logger("DEBUG","getCommand(): Empty type_key was got");
		return NULL;
	}
	if (item->key == NULL) {
		logger("DEBUG","Empty key was got");
		return NULL;
	}

	sprintf(tmp,"getCommand for >%s<", item->key);
	logger("DEBUG",tmp);   

	return item->cmd;
}

char *parse_CLCC(char *clcc)
{
	// There are could be more than one call at time ... but here we will get then one-by-one
	// +CLCC: 1,,4,0,0,"<phone number here>",129,
	char *start ,*ptr;
	int commaCount = 0;

	ptr = clcc;
	while(ptr && commaCount < 5) {
		if (*ptr == ',') {
			 commaCount++;
		}
		ptr++;
	}

	if (ptr == NULL) { // Call was finished
		return "FINISHED";
	}
	if (ptr && *ptr == '"') {
		ptr++;
	}    
	start = ptr;
	while(ptr && *ptr != '"' &&  *ptr != ',') {
		ptr++;
	}
	if (ptr) {
		*ptr = '\0';
	}

	if(start) {
		sprintf(tmp,"parse_CLCC >%s<", start);
		logger("DEBUG",tmp);   
	} else {
		sprintf(tmp,"parse_CLCC >NULL callerID<");
		logger("DEBUG",tmp);   
		return "NO CALLER ID";
	}

	return start;
}
